@@ -1,5 +1,3 @@
-from copy import deepcopy
-
 from six import with_metaclass, iteritems, itervalues, string_types
 
 from yasv.validators import ValidationError, Validator, NotSpecifiedValue
@@ -18,8 +16,12 @@ class Field(object):
         """
         self.validators = []
         self.label = None
-        self.data = NotSpecifiedValue()
+        self.raw_data = NotSpecifiedValue()
+        self._cleaned_data = NotSpecifiedValue()
         self.errors = []
+        self._is_valid = True
+        self._is_validated = False
+        self.name = ''
 
         for arg in args:
             if isinstance(arg, string_types):
@@ -27,6 +29,37 @@ class Field(object):
             elif isinstance(arg, Validator):
                 self.validators.append(arg)
 
+    def __repr__(self):
+        return '<yasv.core.Field object {0}>'.format(self.name)
+
+    @property
+    def cleaned_data(self):
+        if not self._is_validated:
+            self.validate()
+        return self._cleaned_data
+
+    @cleaned_data.setter
+    def cleaned_data(self, value):
+        self._cleaned_data = value
+        self._is_valid = True
+        self._is_validated = True
+
+    @property
+    def is_valid(self):
+        if not self._is_validated:
+            self.validate()
+        return self._is_valid
+
+    def validate(self):
+        if not self._is_validated:
+            self._is_validated = True
+            for validator in self.validators:
+                try:
+                    validator.validate(self, self._schema)
+                except ValidationError as e:
+                    self._is_valid = False
+                    self.errors.append(e.message)
+
 
 class SchemaMeta(type):
     """ The metaclass for `Schema` and any subclasses of `Schema`.
@@ -84,13 +117,20 @@ class Schema(with_metaclass(SchemaMeta)):
         Accepts data as a dict or namedtuple or any object with attributes.
         Creates fields dict with data.
         """
-        self.fields = deepcopy(self._unbound_fields)
+        self._is_valid = True
+        self._is_validated = False
+        self._fields = {}
+        for name, field in iteritems(self._unbound_fields):
+            self._fields[name] = field.__class__(field.label, *field.validators)
+            self._fields[name]._schema = self
+            self._fields[name].name = name
+
         if isinstance(data, dict):
             for name, value in iteritems(data):
                 self._add_data_to_field(name, value)
         else:
             for name in dir(data):
-                if not name.startswith('_') and not name.startswith('__'):
+                if not name.startswith('_'):
                     try:
                         value = getattr(data, name)
                     except AttributeError:
@@ -98,32 +138,55 @@ class Schema(with_metaclass(SchemaMeta)):
                     else:
                         self._add_data_to_field(name, value)
 
+    def __getitem__(self, key):
+        return self._fields[key]
+
+    def __setitem__(self, key, value):
+        self._fields[key] = value
+
+    def __delitem__(self, key):
+        del self._fields[key]
+
+    def __contains__(self, item):
+        return item in self._fields
+
+    def __iter__(self):
+        return iter(self._fields)
+
+    def values(self):
+        for value in itervalues(self._fields):
+            yield value
+
+    def items(self):
+        for key, value in iteritems(self._fields):
+            yield key, value
+
     def _add_data_to_field(self, name, value):
-        if name in self.fields:
-            self.fields[name].data = value
+        if name in self:
+            self[name].raw_data = value
+            self[name]._cleaned_data = value
 
-    def is_valid(self):
-        """ Validate field value.
-
-        Constructs a cleaned_data dict with valid, modified data.
-        Constructs an errors list with error messages.
-        Returns validation status.
+    def validate(self):
+        """ Validate fields values if they are not validated.
+        Set schema validation status.
         """
-        is_valid = True
-        self.cleaned_data = {}
-        for name, field in iteritems(self.fields):
-            for validator in field.validators:
-                try:
-                    value = self.cleaned_data.get(name, None) or field.data
-                    self.cleaned_data[name] = validator.validate(value, field)
-                except ValidationError as e:
-                    is_valid = False
-                    field.errors.append(e.message)
-                    break
+        if not self._is_validated:
+            for field in self.values():
+                if not field.is_valid:
+                    self._is_valid = False
 
-        return is_valid
+    @property
+    def is_valid(self):
+        """ Return schema validation status. Run `validate` if needed.
+        """
+        if not self._is_validated:
+            self.validate()
+        return self._is_valid
 
     def get_errors(self):
-        """ Return a list of error messages.
+        """ Return a dict of field_name: [field_errors].
         """
-        return sum([x.errors for x in itervalues(self.fields)], [])
+        if not self._is_validated:
+            self.validate()
+        return {name: field.errors for name, field in self.items()
+                if field.errors}
